// *******************
// * PM Tools Plugin *
// *******************
// WIP version
// by Pierre-Marie Baty <pm@bots-united.com>


#include <extdll.h>
#include <dllapi.h>
#include <h_export.h>
#include <meta_api.h>


enginefuncs_t g_engfuncs;
globalvars_t  *gpGlobals;
DLL_FUNCTIONS gFunctionTable;
edict_t *pWorldEntity = NULL;
edict_t *pListenserverEntity = NULL;
int beam_texture = 0;
cvar_t hook_dispatch = {"hook_dispatch", "0", FCVAR_EXTDLL};
cvar_t hook_network = {"hook_network", "0", FCVAR_EXTDLL};
cvar_t hook_alert = {"hook_alert", "0", FCVAR_EXTDLL};
cvar_t hook_messages = {"hook_messages", "", FCVAR_EXTDLL};
cvar_t logfile = {"logfile", "", FCVAR_EXTDLL};
cvar_t iaintnocheater = {"iaintnocheater", "0", FCVAR_EXTDLL};
cvar_t iaintnobirdie = {"iaintnobirdie", "0", FCVAR_EXTDLL};
bool hook_this_message = FALSE;

void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals);
int Spawn (edict_t *pent);
void Think (edict_t *pent);
void Use (edict_t *pentUsed, edict_t *pentOther);
void Touch (edict_t *pentTouched, edict_t *pentOther);
void Blocked (edict_t *pentBlocked, edict_t *pentOther);
void KeyValue (edict_t *pentKeyvalue, KeyValueData *pkvd);
void Save (edict_t *pent, SAVERESTOREDATA *pSaveData);
int Restore (edict_t *pent, SAVERESTOREDATA *pSaveData, int globalEntity);
void SetAbsBox (edict_t *pent);
BOOL ClientConnect (edict_t *pEntity, const char *pszName, const char *pszAddress, char szRejectReason[128]);
void ClientDisconnect (edict_t *pEntity);
void StartFrame (void);
void AlertMessage (ALERT_TYPE atype, char *szFmt, ...);
void MessageBegin (int msg_dest, int msg_type, const float *pOrigin, edict_t *ed);
void MessageEnd (void);
void WriteByte (int iValue);
void WriteChar (int iValue);
void WriteShort (int iValue);
void WriteLong (int iValue);
void WriteAngle (float flValue);
void WriteCoord (float flValue);
void WriteString (const char *sz);
void WriteEntity (int iValue);
void ServerCommand_ListEntities (void);
void ServerCommand_PrintEntity (void);
void ServerCommand_SearchEntities (void);
void ServerCommand_SearchEntitiesAll (void);
void ServerCommand_TraceLine (void);
void ServerCommand_TraceHull (void);
void ServerCommand_TraceTexture (void);
void ServerCommand_PointContents (void);
void ServerCommand_Locate (void);
void ServerCommand_Teleport (void);
void ServerCommand_DumpGlobals (void);
void ServerCommand_Extract (void);
void UTIL_DrawBeam (Vector start, Vector end, int life, int width, int noise, int red, int green, int blue, int brightness, int speed);
void UTIL_printf (char *fmt, ...);
void UTIL_ServerConsole_printf (char *fmt, ...);
bool FInViewCone (Vector pOrigin, edict_t *pEdict);
bool CanSee (edict_t *pSeeingEntity, edict_t *pEntityToSee);
char *GameDir (void);

// START of Metamod stuff

enginefuncs_t meta_engfuncs;
gamedll_funcs_t *gpGamedllFuncs;
mutil_funcs_t *gpMetaUtilFuncs;
meta_globals_t *gpMetaGlobals;

META_FUNCTIONS gMetaFunctionTable =
{
   NULL, // pfnGetEntityAPI()
   NULL, // pfnGetEntityAPI_Post()
   GetEntityAPI2, // pfnGetEntityAPI2()
   NULL, // pfnGetEntityAPI2_Post()
   NULL, // pfnGetNewDLLFunctions()
   NULL, // pfnGetNewDLLFunctions_Post()
   GetEngineFunctions, // pfnGetEngineFunctions()
   NULL, // pfnGetEngineFunctions_Post()
};

plugin_info_t Plugin_info = {
   META_INTERFACE_VERSION, // interface version
   "PMTools", // plugin name
   __DATE__, // plugin version
   __DATE__, // date of creation
   "Pierre-Marie Baty <pm@bots-united.com>", // plugin author
   "http://racc.bots-united.com", // plugin URL
   "PMTOOLS", // plugin logtag
   PT_CHANGELEVEL, // when loadable
   PT_ANYTIME, // when unloadable
};


C_DLLEXPORT int Meta_Query (char *ifvers, plugin_info_t **pPlugInfo, mutil_funcs_t *pMetaUtilFuncs)
{
   // this function is the first function ever called by metamod in the plugin DLL. Its purpose
   // is for metamod to retrieve basic information about the plugin, such as its meta-interface
   // version, for ensuring compatibility with the current version of the running metamod.

   // keep track of the pointers to metamod function tables metamod gives us
   gpMetaUtilFuncs = pMetaUtilFuncs;
   *pPlugInfo = &Plugin_info;

   // check for interface version compatibility
   if (strcmp (ifvers, Plugin_info.ifvers) != 0)
   {
      int mmajor = 0, mminor = 0, pmajor = 0, pminor = 0;

      LOG_CONSOLE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);
      LOG_MESSAGE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);

      // if plugin has later interface version, it's incompatible (update metamod)
      sscanf (ifvers, "%d:%d", &mmajor, &mminor);
      sscanf (META_INTERFACE_VERSION, "%d:%d", &pmajor, &pminor);

      if ((pmajor > mmajor) || ((pmajor == mmajor) && (pminor > mminor)))
      {
         LOG_CONSOLE (PLID, "metamod version is too old for this plugin; update metamod");
         LOG_ERROR (PLID, "metamod version is too old for this plugin; update metamod");
         return (FALSE);
      }

      // if plugin has older major interface version, it's incompatible (update plugin)
      else if (pmajor < mmajor)
      {
         LOG_CONSOLE (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         LOG_ERROR (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         return (FALSE);
      }
   }

   return (TRUE); // tell metamod this plugin looks safe
}


C_DLLEXPORT int Meta_Attach (PLUG_LOADTIME now, META_FUNCTIONS *pFunctionTable, meta_globals_t *pMGlobals, gamedll_funcs_t *pGamedllFuncs)
{
   // this function is called when metamod attempts to load the plugin. Since it's the place
   // where we can tell if the plugin will be allowed to run or not, we wait until here to make
   // our initialization stuff, like registering CVARs and dedicated server commands.

   // are we allowed to load this plugin now ?
   if (now > Plugin_info.loadable)
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }

   // keep track of the pointers to engine function tables metamod gives us
   gpMetaGlobals = pMGlobals;
   memcpy (pFunctionTable, &gMetaFunctionTable, sizeof (META_FUNCTIONS));
   gpGamedllFuncs = pGamedllFuncs;

   // print a message to notify about plugin attaching
   LOG_CONSOLE (PLID, "%s: plugin attaching", Plugin_info.name);
   LOG_MESSAGE (PLID, "%s: plugin attaching", Plugin_info.name);

   // ask the engine to register the server commands this plugin uses
   REG_SVR_COMMAND ("listent", ServerCommand_ListEntities);
   REG_SVR_COMMAND ("printent", ServerCommand_PrintEntity);
   REG_SVR_COMMAND ("search", ServerCommand_SearchEntities);
   REG_SVR_COMMAND ("asearch", ServerCommand_SearchEntitiesAll);
   REG_SVR_COMMAND ("traceline", ServerCommand_TraceLine);
   REG_SVR_COMMAND ("tracehull", ServerCommand_TraceHull);
   REG_SVR_COMMAND ("tracetexture", ServerCommand_TraceTexture);
   REG_SVR_COMMAND ("pointcontents", ServerCommand_PointContents);
   REG_SVR_COMMAND ("locate", ServerCommand_Locate);
   REG_SVR_COMMAND ("teleport", ServerCommand_Teleport);
   REG_SVR_COMMAND ("dumpglobals", ServerCommand_DumpGlobals);
   REG_SVR_COMMAND ("extract", ServerCommand_Extract);

   // ask the engine to register the CVARs this plugin uses
   CVAR_REGISTER (&hook_dispatch);
   CVAR_REGISTER (&hook_network);
   CVAR_REGISTER (&hook_alert);
   CVAR_REGISTER (&hook_messages);
   CVAR_REGISTER (&logfile);
   CVAR_REGISTER (&iaintnocheater);
   CVAR_REGISTER (&iaintnobirdie);

   return (TRUE); // returning TRUE enables metamod to attach this plugin
}


C_DLLEXPORT int Meta_Detach (PLUG_LOADTIME now, PL_UNLOAD_REASON reason)
{
   // this function is called when metamod unloads the plugin. A basic check is made in order
   // to prevent unloading the plugin if its processing should not be interrupted.

   // is metamod allowed to unload the plugin ?
   if ((now > Plugin_info.unloadable) && (reason != PNL_CMD_FORCED))
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from unloading this plugin
   }

   return (TRUE); // returning TRUE enables metamod to unload this plugin
}


C_DLLEXPORT int GetEntityAPI2 (DLL_FUNCTIONS *pFunctionTable, int *interfaceVersion)
{
   gFunctionTable.pfnSpawn = Spawn;
   gFunctionTable.pfnThink = Think;
   gFunctionTable.pfnUse = Use;
   gFunctionTable.pfnTouch = Touch;
   gFunctionTable.pfnBlocked = Blocked;
   gFunctionTable.pfnKeyValue = KeyValue;
   gFunctionTable.pfnSave = Save;
   gFunctionTable.pfnRestore = Restore;
   gFunctionTable.pfnSetAbsBox = SetAbsBox;
   gFunctionTable.pfnClientConnect = ClientConnect;
   gFunctionTable.pfnClientDisconnect = ClientDisconnect;
   gFunctionTable.pfnStartFrame = StartFrame;

   memcpy (pFunctionTable, &gFunctionTable, sizeof (DLL_FUNCTIONS));
   return (TRUE);
}


C_DLLEXPORT int GetEngineFunctions (enginefuncs_t *pengfuncsFromEngine, int *interfaceVersion)
{
   meta_engfuncs.pfnAlertMessage = AlertMessage;
   meta_engfuncs.pfnMessageBegin = MessageBegin;
   meta_engfuncs.pfnMessageEnd = MessageEnd;
   meta_engfuncs.pfnWriteByte = WriteByte;
   meta_engfuncs.pfnWriteChar = WriteChar;
   meta_engfuncs.pfnWriteShort = WriteShort;
   meta_engfuncs.pfnWriteLong = WriteLong;
   meta_engfuncs.pfnWriteAngle = WriteAngle;
   meta_engfuncs.pfnWriteCoord = WriteCoord;
   meta_engfuncs.pfnWriteString = WriteString;
   meta_engfuncs.pfnWriteEntity = WriteEntity;

   memcpy (pengfuncsFromEngine, &meta_engfuncs, sizeof (enginefuncs_t));
   return TRUE;
}

// END of Metamod stuff




#ifdef _WIN32
BOOL WINAPI DllMain (HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
   return TRUE; // required Win32 DLL entry point
}
#endif


void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals)
{
   memcpy (&g_engfuncs, pengfuncsFromEngine, sizeof (enginefuncs_t));
   gpGlobals = pGlobals;
}


int Spawn (edict_t *pent)
{
   // do we want to be notified of this function call ?
   if (CVAR_GET_FLOAT ("hook_dispatch") > 1)
      UTIL_printf ("Entity %d (\"%s\") SPAWNS\n", ENTINDEX (pent), (STRING (pent->v.netname) != NULL ? STRING (pent->v.netname) : STRING (pent->v.classname)));

   // if we are spawning the world itself (worldspawn is the first entity spawned)
   if (strcmp (STRING (pent->v.classname), "worldspawn") == 0)
   {
      beam_texture = PRECACHE_MODEL ("sprites/lgtning.spr"); // used to trace beams
      pWorldEntity = pent; // keep track of the world entity
   }

   RETURN_META_VALUE (MRES_IGNORED, 0);
}


void Think (edict_t *pent)
{
   // do we want to be notified of this function call ?
   if (CVAR_GET_FLOAT ("hook_dispatch") > 1)
      UTIL_printf ("Entity %d (\"%s\") THINKS\n", ENTINDEX (pent), (STRING (pent->v.netname) != NULL ? STRING (pent->v.netname) : STRING (pent->v.classname)));

   RETURN_META (MRES_IGNORED);
}


void Use (edict_t *pentUsed, edict_t *pentOther)
{
   // do we want to be notified of this function call ?
   if (CVAR_GET_FLOAT ("hook_dispatch") > 0)
      UTIL_printf ("Entity %d (\"%s\") USES Entity %d (\"%s\")\n", ENTINDEX (pentOther), (STRING (pentOther->v.netname) != NULL ? STRING (pentOther->v.netname) : STRING (pentOther->v.classname)), ENTINDEX (pentUsed), (STRING (pentUsed->v.netname) != NULL ? STRING (pentUsed->v.netname) : STRING (pentUsed->v.classname)));

   RETURN_META (MRES_IGNORED);
}


void Touch (edict_t *pentTouched, edict_t *pentOther)
{
   // do we want to be notified of this function call ?
   if (CVAR_GET_FLOAT ("hook_dispatch") > 0)
      UTIL_printf ("Entity %d (\"%s\") TOUCHES Entity %d (\"%s\")\n", ENTINDEX (pentOther), (STRING (pentOther->v.netname) != NULL ? STRING (pentOther->v.netname) : STRING (pentOther->v.classname)), ENTINDEX (pentTouched), (STRING (pentTouched->v.netname) != NULL ? STRING (pentTouched->v.netname) : STRING (pentTouched->v.classname)));

   RETURN_META (MRES_IGNORED);
}


void Blocked (edict_t *pentBlocked, edict_t *pentOther)
{
   // do we want to be notified of this function call ?
   if (CVAR_GET_FLOAT ("hook_dispatch") > 0)
      UTIL_printf ("Entity %d (\"%s\") BLOCKS Entity %d (\"%s\")\n", ENTINDEX (pentOther), (STRING (pentOther->v.netname) != NULL ? STRING (pentOther->v.netname) : STRING (pentOther->v.classname)), ENTINDEX (pentBlocked), (STRING (pentBlocked->v.netname) != NULL ? STRING (pentBlocked->v.netname) : STRING (pentBlocked->v.classname)));

   RETURN_META (MRES_IGNORED);
}


void KeyValue (edict_t *pentKeyvalue, KeyValueData *pkvd)
{
   // do we want to be notified of this function call ?
   if (CVAR_GET_FLOAT ("hook_dispatch") > 1)
      UTIL_printf ("Entity %d (\"%s\") SETS KEY \"%s\" TO VALUE \"%s\" FOR Classname \"%s\"\n", ENTINDEX (pentKeyvalue), (STRING (pentKeyvalue->v.netname) != NULL ? STRING (pentKeyvalue->v.netname) : STRING (pentKeyvalue->v.classname)), pkvd->szKeyName, pkvd->szValue, pkvd->szClassName);

   RETURN_META (MRES_IGNORED);
}


void Save (edict_t *pent, SAVERESTOREDATA *pSaveData)
{
   // do we want to be notified of this function call ?
   if (CVAR_GET_FLOAT ("hook_dispatch") > 0)
      UTIL_printf ("Entity %d (\"%s\") SAVES DATA\n", ENTINDEX (pent), (STRING (pent->v.netname) != NULL ? STRING (pent->v.netname) : STRING (pent->v.classname)));

   RETURN_META (MRES_IGNORED);
}


int Restore (edict_t *pent, SAVERESTOREDATA *pSaveData, int globalEntity)
{
   // do we want to be notified of this function call ?
   if (CVAR_GET_FLOAT ("hook_dispatch") > 0)
      UTIL_printf ("Entity %d (\"%s\") RESTORES DATA\n", ENTINDEX (pent), (STRING (pent->v.netname) != NULL ? STRING (pent->v.netname) : STRING (pent->v.classname)));

   RETURN_META_VALUE (MRES_IGNORED, 0);
}


void SetAbsBox (edict_t *pent)
{
   // do we want to be notified of this function call ?
   if (CVAR_GET_FLOAT ("hook_dispatch") > 3)
      UTIL_printf ("Entity %d (\"%s\") SETS OBJECT COLLISION BOX\n", ENTINDEX (pent), (STRING (pent->v.netname) != NULL ? STRING (pent->v.netname) : STRING (pent->v.classname)));

   RETURN_META (MRES_IGNORED);
}


BOOL ClientConnect (edict_t *pEntity, const char *pszName, const char *pszAddress, char szRejectReason[128])
{ 
   // is this client the listen server client ?
   if ((strcmp (pszAddress, "loopback") == 0) || (strcmp (pszAddress, "127.0.0.1") == 0))
      pListenserverEntity = pEntity; // save its edict to pListenserverEntity

   // do we want to be notified of this function call ?
   if (CVAR_GET_FLOAT ("hook_network") > 0)
      UTIL_printf ("Client %d (Entity %d) CONNECTS AS \"%s\" FROM %s\n", ENTINDEX (pEntity), ENTINDEX (pEntity), pszName, pszAddress);

   RETURN_META_VALUE (MRES_IGNORED, FALSE);
}


void ClientDisconnect (edict_t *pEntity)
{
   // is the client disconnecting the listen server client ?
   if (pEntity == pListenserverEntity)
      pListenserverEntity = NULL; // if so, reset his entity pointer

   RETURN_META (MRES_IGNORED);
}


void StartFrame (void)
{
   if (FNullEnt (pListenserverEntity))
      RETURN_META (MRES_IGNORED);

   // is the listen server client invincible ?
   if (CVAR_GET_FLOAT ("iaintnocheater") > 0)
      pListenserverEntity->v.takedamage = DAMAGE_NO;
   else if ((pListenserverEntity->v.health > 0) && !(pListenserverEntity->v.flags & FL_SPECTATOR))
      pListenserverEntity->v.takedamage = DAMAGE_AIM;

   // is the listen server client in noclip mode ?
   if (CVAR_GET_FLOAT ("iaintnobirdie") > 0)
      pListenserverEntity->v.movetype = MOVETYPE_NOCLIP;
   else if ((pListenserverEntity->v.health > 0) && !(pListenserverEntity->v.flags & FL_SPECTATOR)
            && (pListenserverEntity->v.movetype == MOVETYPE_NOCLIP))
      pListenserverEntity->v.movetype = MOVETYPE_STEP;

   RETURN_META (MRES_IGNORED);
}


void AlertMessage (ALERT_TYPE atype, char *szFmt, ...)
{
   va_list argptr;
   static char string[1024];

   // if hook_alert is set, send alert messages to console as well
   if (CVAR_GET_FLOAT ("hook_alert") > 0)
   {
      va_start (argptr, szFmt);
      vsprintf (string, szFmt, argptr);
      va_end (argptr);

      UTIL_printf ("ALERT MESSAGE (%s): %s\n", (atype == at_notice ? "at_notice" : 
                                                (atype == at_console ? "at_console" :
                                                 (atype == at_aiconsole ? "at_aiconsole" :
                                                  (atype == at_warning ? "at_warning" :
                                                   (atype == at_error ? "at_error" :
                                                    (atype == at_logged ? "at_logged" : "unknown alert")))))),
                                               string);
   }

   RETURN_META (MRES_IGNORED);
}


void MessageBegin (int msg_dest, int msg_type, const float *pOrigin, edict_t *ed)
{
   char dest[32], origin[32], edict[32];

   hook_this_message = FALSE;

   // do we want to be notified of this function call ?
   if (CVAR_GET_FLOAT ("hook_network") > 0)
   {
      // see if we need to hook this message in depth
      if ((CVAR_GET_FLOAT ("hook_network") > 1)
          && (strstr (CVAR_GET_STRING ("hook_messages"), GET_USER_MSG_NAME (PLID, msg_type, NULL)) != NULL))
         hook_this_message = TRUE;

      // resolve the parameters
      if (msg_dest == MSG_BROADCAST)
         sprintf (dest, "MSG_BROADCAST"); // unreliable to all
      else if (msg_dest == MSG_ONE)
         sprintf (dest, "MSG_ONE"); // reliable to one (msg_entity)
      else if (msg_dest == MSG_ALL)
         sprintf (dest, "MSG_ALL"); // reliable to all
      else if (msg_dest == MSG_INIT)
         sprintf (dest, "MSG_INIT"); // write to the init string
      else if (msg_dest == MSG_PVS)
         sprintf (dest, "MSG_PVS"); // Ents in PVS of org
      else if (msg_dest == MSG_PAS)
         sprintf (dest, "MSG_PAS"); // Ents in PAS of org
      else if (msg_dest == MSG_PVS_R)
         sprintf (dest, "MSG_PVS_R"); // Reliable to PVS
      else if (msg_dest == MSG_PAS_R)
         sprintf (dest, "MSG_PAS_R"); // Reliable to PAS
      else if (msg_dest == MSG_ONE_UNRELIABLE)
         sprintf (dest, "MSG_ONE_UNRELIABLE"); // Send to one client, but don't put in reliable stream, put in unreliable datagram ( could be dropped )
      else if (msg_dest == MSG_SPEC)
         sprintf (dest, "MSG_SPEC"); // Sends to all spectator proxies
      else
         sprintf (dest, "%d [???]", msg_dest);

      if (pOrigin != NULL)
         sprintf (origin, "Vector (%.1f, %.1f, %.1f)", pOrigin[0], pOrigin[1], pOrigin[2]);
      else
         sprintf (origin, "NULL");

      if (!FNullEnt (ed))
         sprintf (edict, "Entity #%d (%s)", ENTINDEX (ed), (ed->v.flags & FL_CLIENT ? STRING (ed->v.netname) : STRING (ed->v.classname)));
      else
         sprintf (edict, "NULLENT [= Game]");

      if (hook_this_message || (CVAR_GET_FLOAT ("hook_network") == 1))
      UTIL_printf ("%s SENDS MESSAGE type %d (\"%s\") in %d (\"%s\") from %s\n", edict, msg_type, GET_USER_MSG_NAME (PLID, msg_type, NULL), msg_dest, dest, origin);

      if (hook_this_message)
         UTIL_ServerConsole_printf ("MESSAGE_BEGIN (%s, %s, %s, %s)\n", dest, GET_USER_MSG_NAME (PLID, msg_type, NULL), origin, edict);
   }

   RETURN_META (MRES_IGNORED);
}


void MessageEnd (void)
{
   // print message end
   if (hook_this_message)
      UTIL_ServerConsole_printf ("MESSAGE_END ()\n");

   hook_this_message = FALSE;

   RETURN_META (MRES_IGNORED);
}


void WriteByte (int iValue)
{
   // print byte
   if (hook_this_message)
      UTIL_ServerConsole_printf ("WRITE_BYTE (%d [0x%x])\n", iValue, iValue);

   RETURN_META (MRES_IGNORED);
}


void WriteChar (int iValue)
{
   // print char
   if (hook_this_message)
      UTIL_ServerConsole_printf ("WRITE_CHAR (%d ['%c'])\n", iValue, iValue);

   RETURN_META (MRES_IGNORED);
}


void WriteShort (int iValue)
{
   // print short
   if (hook_this_message)
      UTIL_ServerConsole_printf ("WRITE_SHORT (%d)\n", iValue);

   RETURN_META (MRES_IGNORED);
}


void WriteLong (int iValue)
{
   // print long
   if (hook_this_message)
      UTIL_ServerConsole_printf ("WRITE_LONG (%d)\n", iValue);

   RETURN_META (MRES_IGNORED);
}


void WriteAngle (float flValue)
{
   // print angle
   if (hook_this_message)
      UTIL_ServerConsole_printf ("WRITE_ANGLE (%.2f)\n", flValue);

   RETURN_META (MRES_IGNORED);
}


void WriteCoord (float flValue)
{
   // print coord
   if (hook_this_message)
      UTIL_ServerConsole_printf ("WRITE_COORD (%.2f)\n", flValue);

   RETURN_META (MRES_IGNORED);
}


void WriteString (const char *sz)
{
   // print string
   if (hook_this_message)
      UTIL_ServerConsole_printf ("WRITE_STRING (\"%s\")\n", sz);

   RETURN_META (MRES_IGNORED);
}


void WriteEntity (int iValue)
{
   // resolve the entity index
   if (hook_this_message)
   {
      char entity_name[32];
      edict_t *pEntity = INDEXENT (iValue);

      if (FNullEnt (pEntity))
         sprintf (entity_name, " [NULLENT]");
      else if (pEntity->v.flags & FL_CLIENT)
         sprintf (entity_name, " [%s]", STRING (pEntity->v.netname));
      else
         sprintf (entity_name, " [%s]", STRING (pEntity->v.classname));

      UTIL_ServerConsole_printf ("WRITE_ENTITY (%d%s)\n", iValue, entity_name);
   }

   RETURN_META (MRES_IGNORED);
}


void ServerCommand_ListEntities (void)
{
   edict_t *pEntity = NULL;
   bool is_player;
   Vector v_origin;
   float f_distance;
   int index = 0;

   char arg1[128]; // no way, I hate pointers...
   
   sprintf (arg1, CMD_ARGV (1)); // get any argument to the command

   // tell people what we are going to do
   UTIL_ServerConsole_printf ("Printing out ALL entities in game...\n");
   UTIL_ServerConsole_printf ("index CLASSNAME [\"netname\"] (model): absmin (x, y, z); size (x, y, z); distance d\n");

   // loop through ALL the entities in the game
   for (index = 0; index < gpGlobals->maxEntities; index++)
   {
      pEntity = INDEXENT (index); // bind that entity number
      if (FNullEnt (pEntity))
         continue; // if unregistered, skip it

      is_player = FALSE; // assume this entity is not a player until we do the check for it

      // if we want a special type of entity and this one does not match...
      if (strstr (STRING (pEntity->v.classname), arg1) == NULL)
         continue; // then skip it

      // is this entity a player ?
      if (strcmp ("player", STRING (pEntity->v.classname)) == 0)
         is_player = TRUE; // if so, remember it

      v_origin = pEntity->v.absmin + (pEntity->v.size * 0.5); // get its origin
      if (FNullEnt (pListenserverEntity)
          || (v_origin == Vector (0, 0, 0)) || (v_origin == Vector (-1, -1, -1)))
         f_distance = -1; // if no origin, then no distance either
      else
         f_distance = (v_origin - pListenserverEntity->v.origin).Length (); // else get the distance to it

      // print a nice line about that entity (with his net name if it's a player)
      if (is_player)
         UTIL_ServerConsole_printf ("%d %s \"%s\" (%s): min (%.0f, %.0f, %.0f); siz (%.0f, %.0f, %.0f); dst %.0f\n", ENTINDEX (pEntity), STRING (pEntity->v.classname), STRING (pEntity->v.netname), g_engfuncs.pfnInfoKeyValue ((*g_engfuncs.pfnGetInfoKeyBuffer) (pEntity), "model"), pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z, pEntity->v.size.x, pEntity->v.size.y, pEntity->v.size.z, f_distance);
      else
         UTIL_ServerConsole_printf ("%d %s (%s): min (%.0f, %.0f, %.0f); siz (%.0f, %.0f, %.0f); dst %.0f\n", ENTINDEX (pEntity), STRING (pEntity->v.classname), STRING (pEntity->v.model), pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z, pEntity->v.size.x, pEntity->v.size.y, pEntity->v.size.z, f_distance);

      // if this entity is visible and has a size...
      if (!FNullEnt (pListenserverEntity) && CanSee (pListenserverEntity, pEntity) && FInViewCone (v_origin, pListenserverEntity) && (pEntity->v.size != Vector (0, 0, 0)))
      {
         // draw a box around it
         UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
         UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
         UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
         UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
         UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
         UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
         UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
         UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
         UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
         UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
         UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
         UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);

         // if it's a player, draw his view offset
         if (is_player)
            UTIL_DrawBeam (v_origin, v_origin + pEntity->v.view_ofs, 2000, 10, 0, 255, 0, 0, 255, 0);
      }
   }

   UTIL_ServerConsole_printf ("End of list - %d entities found.\n", index); // print the number of entities found
   return;
}


void ServerCommand_PrintEntity (void)
{
   edict_t *pEntity = NULL;
   Vector v_origin;
   int index;

   char arg1[128]; // no way, I hate pointers...
   char temp[128];
   
   sprintf (arg1, CMD_ARGV (1)); // get any argument to the command

   // if we don't have any argument, return with an error message
   if (arg1[0] == 0)
   {
      UTIL_ServerConsole_printf ("Usage: \"%s INDEX_OF_ENTITY\"\n", CMD_ARGV (0));
      return; // abort command if no argument
   }

   // try to bind that entity
   index = atoi (arg1);
   if (index == 0)
      pEntity = pWorldEntity; // we asked for the world
   else
      pEntity = INDEXENT (index); // we asked for a normal entity

   // if the entity we want is unregistered, return with an error message
   if ((index != 0) && FNullEnt (pEntity))
   {
      UTIL_ServerConsole_printf ("%s: entity #%d is unregistered\n", CMD_ARGV (0), index);
      return; // abort command if error
   }

   // tell people what we are going to do
   UTIL_ServerConsole_printf ("Printing out entity #%d variable information...\n", ENTINDEX (pEntity));
   UTIL_ServerConsole_printf ("(variable name = value (meaning))\n");

   // now just dump the entity variables
   UTIL_ServerConsole_printf ("pEntity->v.classname = %d (\"%s\")\n", pEntity->v.classname, STRING (pEntity->v.classname));
   UTIL_ServerConsole_printf ("pEntity->v.globalname = %d (\"%s\")\n", pEntity->v.globalname, STRING (pEntity->v.globalname));
   UTIL_ServerConsole_printf ("pEntity->v.origin = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.origin.x, pEntity->v.origin.y, pEntity->v.origin.z);
   UTIL_ServerConsole_printf ("pEntity->v.oldorigin = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.oldorigin.x, pEntity->v.oldorigin.y, pEntity->v.oldorigin.z);
   UTIL_ServerConsole_printf ("pEntity->v.velocity = Vector (%.1f, %.1f, %.1f) - Length %.3f - Length2D %.3f\n", pEntity->v.velocity.x, pEntity->v.velocity.y, pEntity->v.velocity.z, pEntity->v.velocity.Length (), pEntity->v.velocity.Length2D ());
   UTIL_ServerConsole_printf ("pEntity->v.basevelocity = Vector (%.1f, %.1f, %.1f) - Length %.3f - Length2D %.3f\n", pEntity->v.basevelocity.x, pEntity->v.basevelocity.y, pEntity->v.basevelocity.z, pEntity->v.velocity.Length (), pEntity->v.velocity.Length2D ());
   UTIL_ServerConsole_printf ("pEntity->v.clbasevelocity = Vector (%.1f, %.1f, %.1f) - Length %.3f - Length2D %.3f\n", pEntity->v.clbasevelocity.x, pEntity->v.clbasevelocity.y, pEntity->v.clbasevelocity.z, pEntity->v.velocity.Length (), pEntity->v.velocity.Length2D ());
   UTIL_ServerConsole_printf ("pEntity->v.movedir = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.movedir.x, pEntity->v.movedir.y, pEntity->v.movedir.z);
   UTIL_ServerConsole_printf ("pEntity->v.angles = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.angles.x, pEntity->v.angles.y, pEntity->v.angles.z);
   UTIL_ServerConsole_printf ("pEntity->v.avelocity = Vector (%.1f, %.1f, %.1f) - Length %.3f - Length2D %.3f\n", pEntity->v.avelocity.x, pEntity->v.avelocity.y, pEntity->v.avelocity.z, pEntity->v.velocity.Length (), pEntity->v.velocity.Length2D ());
   UTIL_ServerConsole_printf ("pEntity->v.punchangle = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.punchangle.x, pEntity->v.punchangle.y, pEntity->v.punchangle.z);
   UTIL_ServerConsole_printf ("pEntity->v.v_angle = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.v_angle.x, pEntity->v.v_angle.y, pEntity->v.v_angle.z);
   UTIL_ServerConsole_printf ("pEntity->v.endpos = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.endpos.x, pEntity->v.endpos.y, pEntity->v.endpos.z);
   UTIL_ServerConsole_printf ("pEntity->v.startpos = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.startpos.x, pEntity->v.startpos.y, pEntity->v.startpos.z);
   UTIL_ServerConsole_printf ("pEntity->v.impacttime = %.3f\n", pEntity->v.impacttime);
   UTIL_ServerConsole_printf ("pEntity->v.starttime = %.3f\n", pEntity->v.starttime);
   UTIL_ServerConsole_printf ("pEntity->v.fixangle = %d (%s)\n", pEntity->v.fixangle, (pEntity->v.fixangle == 0 ? "nothing" : (pEntity->v.fixangle == 1 ? "force view angles" : (pEntity->v.fixangle == 2 ? "add velocity" : "???"))));
   UTIL_ServerConsole_printf ("pEntity->v.idealpitch = %.3f\n", pEntity->v.idealpitch);
   UTIL_ServerConsole_printf ("pEntity->v.pitch_speed = %.3f\n", pEntity->v.pitch_speed);
   UTIL_ServerConsole_printf ("pEntity->v.ideal_yaw = %.3f\n", pEntity->v.ideal_yaw);
   UTIL_ServerConsole_printf ("pEntity->v.yaw_speed = %.3f\n", pEntity->v.yaw_speed);
   UTIL_ServerConsole_printf ("pEntity->v.modelindex = %d\n", pEntity->v.modelindex);
   UTIL_ServerConsole_printf ("pEntity->v.model = %d (\"%s\")\n", pEntity->v.model, STRING (pEntity->v.model));
   UTIL_ServerConsole_printf ("pEntity->v.viewmodel = %d (\"%s\")\n", pEntity->v.viewmodel, STRING (pEntity->v.viewmodel));
   UTIL_ServerConsole_printf ("pEntity->v.weaponmodel = %d (\"%s\")\n", pEntity->v.weaponmodel, STRING (pEntity->v.weaponmodel));
   UTIL_ServerConsole_printf ("pEntity->v.absmin = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z);
   UTIL_ServerConsole_printf ("pEntity->v.absmax = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.absmax.x, pEntity->v.absmax.y, pEntity->v.absmax.z);
   UTIL_ServerConsole_printf ("pEntity->v.mins = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.mins.x, pEntity->v.mins.y, pEntity->v.mins.z);
   UTIL_ServerConsole_printf ("pEntity->v.maxs = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.maxs.x, pEntity->v.maxs.y, pEntity->v.maxs.z);
   UTIL_ServerConsole_printf ("pEntity->v.size = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.size.x, pEntity->v.size.y, pEntity->v.size.z);
   UTIL_ServerConsole_printf ("pEntity->v.ltime = %.3f (current time is %.3f)\n", pEntity->v.ltime, gpGlobals->time);
   UTIL_ServerConsole_printf ("pEntity->v.nextthink = %.3f (current time is %.3f)\n", pEntity->v.nextthink, gpGlobals->time);
   UTIL_ServerConsole_printf ("pEntity->v.movetype = %d (%s)\n", pEntity->v.movetype, (pEntity->v.movetype == 0 ? "MOVETYPE_NONE" : (pEntity->v.movetype == MOVETYPE_WALK ? "MOVETYPE_WALK" : (pEntity->v.movetype == MOVETYPE_STEP ? "MOVETYPE_STEP" : (pEntity->v.movetype == MOVETYPE_FLY ? "MOVETYPE_FLY" : (pEntity->v.movetype == MOVETYPE_TOSS ? "MOVETYPE_TOSS" : (pEntity->v.movetype == MOVETYPE_PUSH ? "MOVETYPE_PUSH" : (pEntity->v.movetype == MOVETYPE_NOCLIP ? "MOVETYPE_NOCLIP" : (pEntity->v.movetype == MOVETYPE_FLYMISSILE ? "MOVETYPE_FLYMISSILE" : (pEntity->v.movetype == MOVETYPE_BOUNCE ? "MOVETYPE_BOUNCE" : (pEntity->v.movetype == MOVETYPE_BOUNCEMISSILE ? "MOVETYPE_BOUNCEMISSILE" : (pEntity->v.movetype == MOVETYPE_FOLLOW ? "MOVETYPE_FOLLOW" : (pEntity->v.movetype == MOVETYPE_PUSHSTEP ? "MOVETYPE_PUSHSTEP" : "???")))))))))))));
   UTIL_ServerConsole_printf ("pEntity->v.solid = %d (%s)\n", pEntity->v.solid, (pEntity->v.solid == SOLID_NOT ? "SOLID_NOT" : (pEntity->v.solid == SOLID_TRIGGER ? "SOLID_TRIGGER" : (pEntity->v.solid == SOLID_BBOX ? "SOLID_BBOX" : (pEntity->v.solid == SOLID_SLIDEBOX ? "SOLID_SLIDEBOX" : (pEntity->v.solid == SOLID_BSP ? "SOLID_BSP" : "???"))))));
   UTIL_ServerConsole_printf ("pEntity->v.skin = %d (\"%s\")\n", pEntity->v.skin, STRING (pEntity->v.skin));
   UTIL_ServerConsole_printf ("pEntity->v.body = %d (\"%s\")\n", pEntity->v.body, STRING (pEntity->v.body));
   sprintf (temp, ""); // prepare pattern string
   if (pEntity->v.effects & EF_BRIGHTFIELD) strcat (temp, "EF_BRIGHTFIELD | ");
   if (pEntity->v.effects & EF_MUZZLEFLASH) strcat (temp, "EF_MUZZLEFLASH | ");
   if (pEntity->v.effects & EF_BRIGHTLIGHT) strcat (temp, "EF_BRIGHTLIGHT | ");
   if (pEntity->v.effects & EF_DIMLIGHT) strcat (temp, "EF_DIMLIGHT | ");
   if (pEntity->v.effects & EF_INVLIGHT) strcat (temp, "EF_INVLIGHT | ");
   if (pEntity->v.effects & EF_NOINTERP) strcat (temp, "EF_NOINTERP | ");
   if (pEntity->v.effects & EF_LIGHT) strcat (temp, "EF_LIGHT | ");
   if (pEntity->v.effects & EF_NODRAW) strcat (temp, "EF_NODRAW | ");
   if (strlen (temp) > 2) temp[strlen (temp) - 3] = 0; // remove trailing " | "
   UTIL_ServerConsole_printf ("pEntity->v.effects = %d (%s)\n", pEntity->v.effects, temp);
   UTIL_ServerConsole_printf ("pEntity->v.gravity = %.3f (fraction /1 of normal)\n", pEntity->v.gravity);
   UTIL_ServerConsole_printf ("pEntity->v.friction = %.3f\n", pEntity->v.friction);
   UTIL_ServerConsole_printf ("pEntity->v.light_level = %d\n", pEntity->v.light_level);
   UTIL_ServerConsole_printf ("pEntity->v.sequence = %d (\"%s\")\n", pEntity->v.sequence, STRING (pEntity->v.sequence));
   UTIL_ServerConsole_printf ("pEntity->v.gaitsequence = %d (\"%s\")\n", pEntity->v.gaitsequence, STRING (pEntity->v.gaitsequence));
   UTIL_ServerConsole_printf ("pEntity->v.frame = %.3f (position /255 of total in sequence)\n", pEntity->v.frame);
   UTIL_ServerConsole_printf ("pEntity->v.animtime = %.3f (current time is %.3f)\n", pEntity->v.animtime, gpGlobals->time);
   UTIL_ServerConsole_printf ("pEntity->v.framerate = %.3f (times normal speed)\n", pEntity->v.framerate);
   UTIL_ServerConsole_printf ("pEntity->v.controller = {%d, %d, %d, %d}\n", pEntity->v.controller[0], pEntity->v.controller[1], pEntity->v.controller[2], pEntity->v.controller[3]);
   UTIL_ServerConsole_printf ("pEntity->v.blending = {%d, %d}\n", pEntity->v.blending[0], pEntity->v.blending[1]);
   UTIL_ServerConsole_printf ("pEntity->v.scale = %.3f\n", pEntity->v.scale);
   UTIL_ServerConsole_printf ("pEntity->v.rendermode = %d (%s)\n", pEntity->v.rendermode, (pEntity->v.rendermode == kRenderNormal ? "kRenderNormal" : (pEntity->v.rendermode == kRenderTransColor ? "kRenderTransColor" : (pEntity->v.rendermode == kRenderTransTexture ? "kRenderTransTexture" : (pEntity->v.rendermode == kRenderGlow ? "kRenderGlow" : (pEntity->v.rendermode == kRenderTransAlpha ? "kRenderTransAlpha" : (pEntity->v.rendermode == kRenderTransAdd ? "kRenderTransAdd" : "???")))))));
   UTIL_ServerConsole_printf ("pEntity->v.renderamt = %.3f\n", pEntity->v.renderamt);
   UTIL_ServerConsole_printf ("pEntity->v.rendercolor = Vector (%.1f, %.1f, %.1f) (RGB)\n", pEntity->v.rendercolor.x, pEntity->v.rendercolor.y, pEntity->v.rendercolor.z);
   UTIL_ServerConsole_printf ("pEntity->v.renderfx = %d (%s)\n", pEntity->v.renderfx, (pEntity->v.renderfx == kRenderFxNone ? "kRenderFxNone" : (pEntity->v.renderfx == kRenderFxPulseSlow ? "kRenderFxPulseSlow" : (pEntity->v.renderfx == kRenderFxPulseFast ? "kRenderFxPulseFast" : (pEntity->v.renderfx == kRenderFxPulseSlowWide ? "kRenderFxPulseSlowWide" : (pEntity->v.renderfx == kRenderFxPulseFastWide ? "kRenderFxPulseFastWide" : (pEntity->v.renderfx == kRenderFxFadeSlow ? "kRenderFxFadeSlow" : (pEntity->v.renderfx == kRenderFxFadeFast ? "kRenderFxFadeFast" : (pEntity->v.renderfx == kRenderFxSolidSlow ? "kRenderFxSolidSlow" : (pEntity->v.renderfx == kRenderFxSolidFast ? "kRenderFxSolidFast" : (pEntity->v.renderfx == kRenderFxStrobeSlow ? "kRenderFxStrobeSlow" : (pEntity->v.renderfx == kRenderFxStrobeFast ? "kRenderFxStrobeFast" : (pEntity->v.renderfx == kRenderFxStrobeFaster ? "kRenderFxStrobeFaster" : (pEntity->v.renderfx == kRenderFxFlickerSlow ? "kRenderFxFlickerSlow" : (pEntity->v.renderfx == kRenderFxFlickerFast ? "kRenderFxFlickerFast" : (pEntity->v.renderfx == kRenderFxNoDissipation ? "kRenderFxNoDissipation" : (pEntity->v.renderfx == kRenderFxDistort ? "kRenderFxDistort" : (pEntity->v.renderfx == kRenderFxHologram ? "kRenderFxHologram" : (pEntity->v.renderfx == kRenderFxDeadPlayer ? "kRenderFxDeadPlayer" : (pEntity->v.renderfx == kRenderFxExplode ? "kRenderFxExplode" : (pEntity->v.renderfx == kRenderFxGlowShell ? "kRenderFxGlowShell" : (pEntity->v.renderfx == kRenderFxClampMinScale ? "kRenderFxClampMinScale" : "???"))))))))))))))))))))));
   UTIL_ServerConsole_printf ("pEntity->v.health = %.3f\n", pEntity->v.health);
   UTIL_ServerConsole_printf ("pEntity->v.frags = %.3f\n", pEntity->v.frags);
   sprintf (temp, "00000000000000000000000000000000"); // prepare pattern string
   for (index = 0; index < 32; index++) if (pEntity->v.weapons & (1 << index)) temp[index] = '1';
   UTIL_ServerConsole_printf ("pEntity->v.weapons = %d (%s)\n", pEntity->v.weapons, temp);
   UTIL_ServerConsole_printf ("pEntity->v.takedamage = %.3f (%s)\n", pEntity->v.takedamage, (pEntity->v.takedamage == DAMAGE_NO ? "DAMAGE_NO" : (pEntity->v.takedamage == DAMAGE_YES ? "DAMAGE_YES" : (pEntity->v.takedamage == DAMAGE_AIM ? "DAMAGE_AIM" : "???"))));
   UTIL_ServerConsole_printf ("pEntity->v.deadflag = %d (%s)\n", pEntity->v.deadflag, (pEntity->v.deadflag == DEAD_NO ? "DEAD_NO" : (pEntity->v.deadflag == DEAD_DYING ? "DEAD_DYING" : (pEntity->v.deadflag == DEAD_DEAD ? "DEAD_DEAD" : (pEntity->v.deadflag == DEAD_RESPAWNABLE ? "DEAD_RESPAWNABLE" : (pEntity->v.deadflag == DEAD_DISCARDBODY ? "DEAD_DISCARDBODY" : "???"))))));
   UTIL_ServerConsole_printf ("pEntity->v.view_ofs = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.view_ofs.x, pEntity->v.view_ofs.y, pEntity->v.view_ofs.z);
   sprintf (temp, ""); // prepare pattern string
   if (pEntity->v.button & IN_ATTACK) strcat (temp, "IN_ATTACK | ");
   if (pEntity->v.button & IN_JUMP) strcat (temp, "IN_JUMP | ");
   if (pEntity->v.button & IN_DUCK) strcat (temp, "IN_DUCK | ");
   if (pEntity->v.button & IN_FORWARD) strcat (temp, "IN_FORWARD | ");
   if (pEntity->v.button & IN_BACK) strcat (temp, "IN_BACK | ");
   if (pEntity->v.button & IN_USE) strcat (temp, "IN_USE | ");
   if (pEntity->v.button & IN_CANCEL) strcat (temp, "IN_CANCEL | ");
   if (pEntity->v.button & IN_LEFT) strcat (temp, "IN_LEFT | ");
   if (pEntity->v.button & IN_RIGHT) strcat (temp, "IN_RIGHT | ");
   if (pEntity->v.button & IN_MOVELEFT) strcat (temp, "IN_MOVELEFT | ");
   if (pEntity->v.button & IN_MOVERIGHT) strcat (temp, "IN_MOVERIGHT | ");
   if (pEntity->v.button & IN_ATTACK2) strcat (temp, "IN_ATTACK2 | ");
   if (pEntity->v.button & IN_RUN) strcat (temp, "IN_RUN | ");
   if (pEntity->v.button & IN_RELOAD) strcat (temp, "IN_RELOAD | ");
   if (pEntity->v.button & IN_ALT1) strcat (temp, "IN_ALT1 | ");
   if (pEntity->v.button & IN_SCORE) strcat (temp, "IN_SCORE | ");
   if (strlen (temp) > 2) temp[strlen (temp) - 3] = 0; // remove trailing " | "
   UTIL_ServerConsole_printf ("pEntity->v.button = %d (%s)\n", pEntity->v.button, temp);
   UTIL_ServerConsole_printf ("pEntity->v.impulse = %d\n", pEntity->v.impulse);
   if (pEntity->v.chain != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.chain), STRING (pEntity->v.chain->v.classname)); else sprintf (temp, "NULL");
   UTIL_ServerConsole_printf ("pEntity->v.chain = %s\n", temp);
   if (pEntity->v.dmg_inflictor != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.dmg_inflictor), STRING (pEntity->v.dmg_inflictor->v.classname)); else sprintf (temp, "NULL");
   UTIL_ServerConsole_printf ("pEntity->v.dmg_inflictor = %s\n", temp);
   if (pEntity->v.enemy != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.enemy), STRING (pEntity->v.enemy->v.classname)); else sprintf (temp, "NULL");
   UTIL_ServerConsole_printf ("pEntity->v.enemy = %s\n", temp);
   if (pEntity->v.aiment != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.aiment), STRING (pEntity->v.aiment->v.classname)); else sprintf (temp, "NULL");
   UTIL_ServerConsole_printf ("pEntity->v.aiment = %s\n", temp);
   if (pEntity->v.owner != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.owner), STRING (pEntity->v.owner->v.classname)); else sprintf (temp, "NULL");
   UTIL_ServerConsole_printf ("pEntity->v.owner = %s\n", temp);
   if (pEntity->v.groundentity != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.groundentity), STRING (pEntity->v.groundentity->v.classname)); else sprintf (temp, "NULL");
   UTIL_ServerConsole_printf ("pEntity->v.groundentity = %s\n", temp);
   sprintf (temp, "00000000000000000000000000000000"); // prepare pattern string
   for (index = 0; index < 32; index++) if (pEntity->v.spawnflags & (1 << index)) temp[index] = '1';
   UTIL_ServerConsole_printf ("pEntity->v.spawnflags = %d (%s)\n", pEntity->v.spawnflags, temp);
   sprintf (temp, ""); // prepare pattern string
   if (pEntity->v.flags & FL_FLY) strcat (temp, "FL_FLY | ");
   if (pEntity->v.flags & FL_SWIM) strcat (temp, "FL_SWIM | ");
   if (pEntity->v.flags & FL_CONVEYOR) strcat (temp, "FL_CONVEYOR | ");
   if (pEntity->v.flags & FL_CLIENT) strcat (temp, "FL_CLIENT | ");
   if (pEntity->v.flags & FL_INWATER) strcat (temp, "FL_INWATER | ");
   if (pEntity->v.flags & FL_MONSTER) strcat (temp, "FL_MONSTER | ");
   if (pEntity->v.flags & FL_GODMODE) strcat (temp, "FL_GODMODE | ");
   if (pEntity->v.flags & FL_NOTARGET) strcat (temp, "FL_NOTARGET | ");
   if (pEntity->v.flags & FL_SKIPLOCALHOST) strcat (temp, "FL_SKIPLOCALHOST | ");
   if (pEntity->v.flags & FL_ONGROUND) strcat (temp, "FL_ONGROUND | ");
   if (pEntity->v.flags & FL_PARTIALGROUND) strcat (temp, "FL_PARTIALGROUND | ");
   if (pEntity->v.flags & FL_WATERJUMP) strcat (temp, "FL_WATERJUMP | ");
   if (pEntity->v.flags & FL_FROZEN) strcat (temp, "FL_FROZEN | ");
   if (pEntity->v.flags & FL_FAKECLIENT) strcat (temp, "FL_FAKECLIENT | ");
   if (pEntity->v.flags & FL_DUCKING) strcat (temp, "FL_DUCKING | ");
   if (pEntity->v.flags & FL_FLOAT) strcat (temp, "FL_FLOAT | ");
   if (pEntity->v.flags & FL_GRAPHED) strcat (temp, "FL_GRAPHED | ");
   if (pEntity->v.flags & FL_IMMUNE_WATER) strcat (temp, "FL_IMMUNE_WATER | ");
   if (pEntity->v.flags & FL_IMMUNE_SLIME) strcat (temp, "FL_IMMUNE_SLIME | ");
   if (pEntity->v.flags & FL_IMMUNE_LAVA) strcat (temp, "FL_IMMUNE_LAVA | ");
   if (pEntity->v.flags & FL_PROXY) strcat (temp, "FL_PROXY | ");
   if (pEntity->v.flags & FL_ALWAYSTHINK) strcat (temp, "FL_ALWAYSTHINK | ");
   if (pEntity->v.flags & FL_BASEVELOCITY) strcat (temp, "FL_BASEVELOCITY | ");
   if (pEntity->v.flags & FL_MONSTERCLIP) strcat (temp, "FL_MONSTERCLIP | ");
   if (pEntity->v.flags & FL_ONTRAIN) strcat (temp, "FL_ONTRAIN | ");
   if (pEntity->v.flags & FL_WORLDBRUSH) strcat (temp, "FL_WORLDBRUSH | ");
   if (pEntity->v.flags & FL_SPECTATOR) strcat (temp, "FL_SPECTATOR | ");
   if (pEntity->v.flags & (1 << 27)) strcat (temp, "FL_THIRDPARTYBOT | ");
   if (pEntity->v.flags & FL_CUSTOMENTITY) strcat (temp, "FL_CUSTOMENTITY | ");
   if (pEntity->v.flags & FL_KILLME) strcat (temp, "FL_KILLME | ");
   if (pEntity->v.flags & FL_DORMANT) strcat (temp, "FL_DORMANT | ");
   if (strlen (temp) > 2) temp[strlen (temp) - 3] = 0; // remove trailing " | "
   UTIL_ServerConsole_printf ("pEntity->v.flags = %d (%s)\n", pEntity->v.flags, temp);
   UTIL_ServerConsole_printf ("pEntity->v.colormap = %d (0x%X)\n", pEntity->v.colormap, pEntity->v.colormap);
   UTIL_ServerConsole_printf ("pEntity->v.team = %d\n", pEntity->v.team);
   UTIL_ServerConsole_printf ("pEntity->v.max_health = %.3f\n", pEntity->v.max_health);
   UTIL_ServerConsole_printf ("pEntity->v.teleport_time = %.3f\n", pEntity->v.teleport_time);
   UTIL_ServerConsole_printf ("pEntity->v.armortype = %.3f\n", pEntity->v.armortype);
   UTIL_ServerConsole_printf ("pEntity->v.armorvalue = %.3f\n", pEntity->v.armorvalue);
   UTIL_ServerConsole_printf ("pEntity->v.waterlevel = %d (%s)\n", pEntity->v.waterlevel, (pEntity->v.waterlevel == 0 ? "not in water" : (pEntity->v.waterlevel == 2 ? "walking in water" : (pEntity->v.waterlevel == 3 ? "swimming in water" : "???"))));
   UTIL_ServerConsole_printf ("pEntity->v.watertype = %d\n", pEntity->v.watertype);
   UTIL_ServerConsole_printf ("pEntity->v.target = %d (\"%s\")\n", pEntity->v.target, STRING (pEntity->v.target));
   UTIL_ServerConsole_printf ("pEntity->v.targetname = %d (\"%s\")\n", pEntity->v.targetname, STRING (pEntity->v.targetname));
   UTIL_ServerConsole_printf ("pEntity->v.netname = %d (\"%s\")\n", pEntity->v.netname, STRING (pEntity->v.netname));
   UTIL_ServerConsole_printf ("pEntity->v.message = %d (\"%s\")\n", pEntity->v.message, STRING (pEntity->v.message));
   UTIL_ServerConsole_printf ("pEntity->v.dmg_take = %.3f\n", pEntity->v.dmg_take);
   UTIL_ServerConsole_printf ("pEntity->v.dmg_save = %.3f\n", pEntity->v.dmg_save);
   UTIL_ServerConsole_printf ("pEntity->v.dmg = %.3f\n", pEntity->v.dmg);
   UTIL_ServerConsole_printf ("pEntity->v.dmgtime = %.3f (current time is %.3f)\n", pEntity->v.dmgtime, gpGlobals->time);
   UTIL_ServerConsole_printf ("pEntity->v.noise = %d (\"%s\")\n", pEntity->v.noise, STRING (pEntity->v.noise));
   UTIL_ServerConsole_printf ("pEntity->v.noise1 = %d (\"%s\")\n", pEntity->v.noise1, STRING (pEntity->v.noise1));
   UTIL_ServerConsole_printf ("pEntity->v.noise2 = %d (\"%s\")\n", pEntity->v.noise2, STRING (pEntity->v.noise2));
   UTIL_ServerConsole_printf ("pEntity->v.noise3 = %d (\"%s\")\n", pEntity->v.noise3, STRING (pEntity->v.noise3));
   UTIL_ServerConsole_printf ("pEntity->v.speed = %.3f\n", pEntity->v.speed);
   UTIL_ServerConsole_printf ("pEntity->v.air_finished = %.3f\n", pEntity->v.air_finished);
   UTIL_ServerConsole_printf ("pEntity->v.pain_finished = %.3f\n", pEntity->v.pain_finished);
   UTIL_ServerConsole_printf ("pEntity->v.radsuit_finished = %.3f\n", pEntity->v.radsuit_finished);
   if (pEntity->v.pContainingEntity != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.pContainingEntity), STRING (pEntity->v.pContainingEntity->v.classname)); else sprintf (temp, "NULL");
   UTIL_ServerConsole_printf ("pEntity->v.pContainingEntity = %s\n", temp);
   UTIL_ServerConsole_printf ("pEntity->v.playerclass = %d\n", pEntity->v.playerclass);
   UTIL_ServerConsole_printf ("pEntity->v.maxspeed = %.3f\n", pEntity->v.maxspeed);
   UTIL_ServerConsole_printf ("pEntity->v.fov = %.3f\n", pEntity->v.fov);
   UTIL_ServerConsole_printf ("pEntity->v.weaponanim = %d\n", pEntity->v.weaponanim);
   UTIL_ServerConsole_printf ("pEntity->v.pushmsec = %d\n", pEntity->v.pushmsec);
   UTIL_ServerConsole_printf ("pEntity->v.bInDuck = %d (%s)\n", pEntity->v.bInDuck, (pEntity->v.bInDuck > 0 ? "TRUE" : "FALSE"));
   UTIL_ServerConsole_printf ("pEntity->v.flTimeStepSound = %d (current time is %.3f)\n", pEntity->v.flTimeStepSound, gpGlobals->time);
   UTIL_ServerConsole_printf ("pEntity->v.flSwimTime = %d (current time is %.3f)\n", pEntity->v.flSwimTime, gpGlobals->time);
   UTIL_ServerConsole_printf ("pEntity->v.flDuckTime = %d (current time is %.3f)\n", pEntity->v.flDuckTime, gpGlobals->time);
   UTIL_ServerConsole_printf ("pEntity->v.iStepLeft = %d\n", pEntity->v.iStepLeft);
   UTIL_ServerConsole_printf ("pEntity->v.flFallVelocity = %.3f\n", pEntity->v.flFallVelocity);
   UTIL_ServerConsole_printf ("pEntity->v.gamestate = %d\n", pEntity->v.gamestate);
   UTIL_ServerConsole_printf ("pEntity->v.oldbuttons = %d\n", pEntity->v.oldbuttons);
   UTIL_ServerConsole_printf ("pEntity->v.groupinfo = %d\n", pEntity->v.groupinfo);
   UTIL_ServerConsole_printf ("pEntity->v.iuser1 = %d\n", pEntity->v.iuser1);
   UTIL_ServerConsole_printf ("pEntity->v.iuser2 = %d\n", pEntity->v.iuser2);
   UTIL_ServerConsole_printf ("pEntity->v.iuser3 = %d\n", pEntity->v.iuser3);
   UTIL_ServerConsole_printf ("pEntity->v.iuser4 = %d\n", pEntity->v.iuser4);
   UTIL_ServerConsole_printf ("pEntity->v.fuser1 = %.3f\n", pEntity->v.fuser1);
   UTIL_ServerConsole_printf ("pEntity->v.fuser2 = %.3f\n", pEntity->v.fuser2);
   UTIL_ServerConsole_printf ("pEntity->v.fuser3 = %.3f\n", pEntity->v.fuser3);
   UTIL_ServerConsole_printf ("pEntity->v.fuser4 = %.3f\n", pEntity->v.fuser4);
   UTIL_ServerConsole_printf ("pEntity->v.vuser1 = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.vuser1.x, pEntity->v.vuser1.y, pEntity->v.vuser1.z);
   UTIL_ServerConsole_printf ("pEntity->v.vuser2 = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.vuser2.x, pEntity->v.vuser2.y, pEntity->v.vuser2.z);
   UTIL_ServerConsole_printf ("pEntity->v.vuser3 = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.vuser3.x, pEntity->v.vuser3.y, pEntity->v.vuser3.z);
   UTIL_ServerConsole_printf ("pEntity->v.vuser4 = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.vuser4.x, pEntity->v.vuser4.y, pEntity->v.vuser4.z);
   if (pEntity->v.euser1 != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.euser1), STRING (pEntity->v.euser1->v.classname)); else sprintf (temp, "NULL");
   UTIL_ServerConsole_printf ("pEntity->v.euser1 = %s\n", temp);
   if (pEntity->v.euser2 != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.euser2), STRING (pEntity->v.euser2->v.classname)); else sprintf (temp, "NULL");
   UTIL_ServerConsole_printf ("pEntity->v.euser2 = %s\n", temp);
   if (pEntity->v.euser3 != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.euser3), STRING (pEntity->v.euser3->v.classname)); else sprintf (temp, "NULL");
   UTIL_ServerConsole_printf ("pEntity->v.euser3 = %s\n", temp);
   if (pEntity->v.euser4 != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.euser4), STRING (pEntity->v.euser4->v.classname)); else sprintf (temp, "NULL");
   UTIL_ServerConsole_printf ("pEntity->v.euser4 = %s\n", temp);

   // if this entity is visible and has a size...
   if (!FNullEnt (pListenserverEntity) && CanSee (pListenserverEntity, pEntity) && FInViewCone (pEntity->v.absmin + (pEntity->v.size * 0.5), pListenserverEntity) && (pEntity->v.size != Vector (0, 0, 0)))
   {
      // draw a box around it
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);

      // if it's a player, draw his view offset
      if (strcmp (STRING (pEntity->v.classname), "player") == 0)
         UTIL_DrawBeam (pEntity->v.absmin + (pEntity->v.size * 0.5), pEntity->v.absmin + (pEntity->v.size * 0.5) + pEntity->v.view_ofs, 2000, 10, 0, 255, 0, 0, 255, 0);
   }

   UTIL_ServerConsole_printf ("End of dump.\n"); // print the conclusion message
   return;
}


void ServerCommand_SearchEntities (void)
{
   edict_t *pEntity = NULL;
   bool is_player;
   Vector v_origin;
   float f_distance;
   float radius = 500;
   int entity_count = 0;
   char arg1[128]; // no way, I hate pointers...

   // if no local client
   if (FNullEnt (pListenserverEntity))
   {
      UTIL_ServerConsole_printf ("%s: only works from a local client console\n", CMD_ARGV (0));
      return;
   }

   sprintf (arg1, CMD_ARGV (1)); // get any argument to the command

   // was a radius specified ?
   if ((atoi (arg1) > 0) && (atoi (arg1) <= 10000))
      radius = atoi (arg1); // if so, then use it

   // tell people what we are going to do
   UTIL_ServerConsole_printf ("Searching for VISIBLE entities in player's neighbourhood (radius %.1f)...\n", radius);
   UTIL_ServerConsole_printf ("index CLASSNAME [\"netname\"] (model): absmin (x, y, z); size (x, y, z); distance d\n");

   while ((pEntity = FIND_ENTITY_IN_SPHERE (pEntity, pListenserverEntity->v.origin, radius)) != NULL)
   {
      if (FNullEnt (pEntity))
         break; // break when the engine reports no entity left

      is_player = FALSE; // assume this entity is not a player until we do the check for it

      // is this entity a player ?
      if (strcmp ("player", STRING (pEntity->v.classname)) == 0)
         is_player = TRUE; // if so, remember it

      v_origin = pEntity->v.absmin + (pEntity->v.size * 0.5); // get its origin
      if (FNullEnt (pListenserverEntity) || (v_origin == Vector (0, 0, 0)) || (v_origin == Vector (-1, -1, -1)))
         f_distance = -1; // if no origin, then no distance either
      else
         f_distance = (v_origin - pListenserverEntity->v.origin).Length (); // else get the distance to it

      // if this entity is not visible...
      if (!CanSee (pListenserverEntity, pEntity) || !FInViewCone (v_origin, pListenserverEntity))
         continue; // then skip it

      // print a nice line about that entity (with his net name if it's a player)
      if (is_player)
         UTIL_ServerConsole_printf ("%d %s \"%s\" (%s): min (%.0f, %.0f, %.0f); siz (%.0f, %.0f, %.0f); dst %.0f\n", ENTINDEX (pEntity), STRING (pEntity->v.classname), STRING (pEntity->v.netname), g_engfuncs.pfnInfoKeyValue ((*g_engfuncs.pfnGetInfoKeyBuffer) (pEntity), "model"), pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z, pEntity->v.size.x, pEntity->v.size.y, pEntity->v.size.z, f_distance);
      else
         UTIL_ServerConsole_printf ("%d %s (%s): min (%.0f, %.0f, %.0f); siz (%.0f, %.0f, %.0f); dst %.0f\n", ENTINDEX (pEntity), STRING (pEntity->v.classname), STRING (pEntity->v.model), pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z, pEntity->v.size.x, pEntity->v.size.y, pEntity->v.size.z, f_distance);

      // draw a box around it
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);

      // if it's a player, draw his view offset
      if (is_player)
         UTIL_DrawBeam (v_origin, v_origin + pEntity->v.view_ofs, 2000, 10, 0, 255, 0, 0, 255, 0);

      entity_count++; // increment the entity count
   }

   UTIL_ServerConsole_printf ("End of list - %d entities found.\n", entity_count); // print the number of entities found
   return;
}


void ServerCommand_SearchEntitiesAll (void)
{
   edict_t *pEntity = NULL;
   bool is_player;
   Vector v_origin;
   float f_distance;
   float radius = 500;
   int entity_count = 0;
   char arg1[128]; // no way, I hate pointers...

   // if no local client
   if (FNullEnt (pListenserverEntity))
   {
      UTIL_ServerConsole_printf ("%s: only works from a local client console\n", CMD_ARGV (0));
      return;
   }

   sprintf (arg1, CMD_ARGV (1)); // get any argument to the command

   // was a radius specified ?
   if ((atoi (arg1) > 0) && (atoi (arg1) <= 10000))
      radius = atoi (arg1); // if so, then use it

   // tell people what we are going to do
   UTIL_ServerConsole_printf ("Searching for ALL entities in player's neighbourhood (radius %.1f)...\n", radius);
   UTIL_ServerConsole_printf ("index CLASSNAME [\"netname\"] (model): absmin (x, y, z); size (x, y, z); distance d\n");

   while ((pEntity = FIND_ENTITY_IN_SPHERE (pEntity, pListenserverEntity->v.origin, radius)) != NULL)
   {
      if (FNullEnt (pEntity))
         break; // break when the engine reports no entity left

      is_player = FALSE; // assume this entity is not a player until we do the check for it

      // is this entity a player ?
      if (strcmp ("player", STRING (pEntity->v.classname)) == 0)
         is_player = TRUE; // if so, remember it

      v_origin = pEntity->v.absmin + (pEntity->v.size * 0.5); // get its origin
      if ((v_origin == Vector (0, 0, 0)) || (v_origin == Vector (-1, -1, -1)))
         f_distance = -1; // if no origin, then no distance either
      else
         f_distance = (v_origin - pListenserverEntity->v.origin).Length (); // else get the distance to it

      // print a nice line about that entity (with his net name if it's a player)
      if (is_player)
         UTIL_ServerConsole_printf ("%d %s \"%s\" (%s): min (%.0f, %.0f, %.0f); siz (%.0f, %.0f, %.0f); dst %.0f\n", ENTINDEX (pEntity), STRING (pEntity->v.classname), STRING (pEntity->v.netname), g_engfuncs.pfnInfoKeyValue ((*g_engfuncs.pfnGetInfoKeyBuffer) (pEntity), "model"), pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z, pEntity->v.size.x, pEntity->v.size.y, pEntity->v.size.z, f_distance);
      else
         UTIL_ServerConsole_printf ("%d %s (%s): min (%.0f, %.0f, %.0f); siz (%.0f, %.0f, %.0f); dst %.0f\n", ENTINDEX (pEntity), STRING (pEntity->v.classname), STRING (pEntity->v.model), pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z, pEntity->v.size.x, pEntity->v.size.y, pEntity->v.size.z, f_distance);

      // draw a box around it
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x + pEntity->v.size.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);
      UTIL_DrawBeam (Vector (pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z + pEntity->v.size.z), Vector (pEntity->v.absmin.x, pEntity->v.absmin.y + pEntity->v.size.y, pEntity->v.absmin.z + pEntity->v.size.z), 2000, 10, 0, 0, 255, 0, 255, 0);

      // if it's a player, draw his view offset
      if (is_player)
         UTIL_DrawBeam (v_origin, v_origin + pEntity->v.view_ofs, 2000, 10, 0, 255, 0, 0, 255, 0);

      entity_count++; // increment the entity count
   }

   UTIL_ServerConsole_printf ("End of list - %d entities found.\n", entity_count); // print the number of entities found
   return;
}


void ServerCommand_TraceLine (void)
{
   float distance = 10000;
   TraceResult tr;
   char args[128][3]; // no way, I hate pointers...
   int ignore_value = 0;
   int arg_num;

   // if no local client
   if (FNullEnt (pListenserverEntity))
   {
      UTIL_ServerConsole_printf ("%s: only works from a local client console\n", CMD_ARGV (0));
      return;
   }

   UTIL_ServerConsole_printf ("Tracing a LINE %.0f units forward from player's view angles (%.1f %.1f %.1f)\n", distance, pListenserverEntity->v.v_angle.x, pListenserverEntity->v.v_angle.y, pListenserverEntity->v.v_angle.z);

   // record arguments
   for (arg_num = 0; arg_num < 3; arg_num++)
      sprintf (args[arg_num], CMD_ARGV (arg_num + 1));

   // parse arguments
   for (arg_num = 0; arg_num < 3; arg_num++)
   {
      if ((atoi (args[arg_num]) > 0) && (atoi (args[arg_num]) <= 10000))
         distance = atoi (args[arg_num]);
      else if (strcmp ("ignore_monsters", args[arg_num]) == 0)
      {
         ignore_value |= 0x0001;
         UTIL_ServerConsole_printf ("IGNORING monsters\n");
      }
      else if (strcmp ("ignore_glass", args[arg_num]) == 0)
      {
         ignore_value |= 0x0101;
         UTIL_ServerConsole_printf ("IGNORING glass\n");
      }
   }

   MAKE_VECTORS (pListenserverEntity->v.v_angle);
   TRACE_LINE (pListenserverEntity->v.origin + pListenserverEntity->v.view_ofs,
               pListenserverEntity->v.origin + pListenserverEntity->v.view_ofs + gpGlobals->v_forward * distance,
               ignore_value, pListenserverEntity, &tr);

   UTIL_DrawBeam (pListenserverEntity->v.origin + pListenserverEntity->v.view_ofs, pListenserverEntity->v.origin + pListenserverEntity->v.view_ofs + gpGlobals->v_forward * distance * tr.flFraction, 600, 10, 2, 250, 250, 250, 200, 10);

   UTIL_ServerConsole_printf ("Detailed trace results:\n");
   UTIL_ServerConsole_printf ("   fAllSolid = %d (plane is %svalid)\n", tr.fAllSolid, (tr.fAllSolid > 0 ? "NOT " : ""));
   UTIL_ServerConsole_printf ("   fStartSolid = %d (initial point was %sin a solid area)\n", tr.fStartSolid, (tr.fStartSolid > 0 ? "" : "NOT "));
   UTIL_ServerConsole_printf ("   fInOpen = %d (%sin open space)\n", tr.fInOpen, (tr.fInOpen > 0 ? "" : "NOT "));
   UTIL_ServerConsole_printf ("   fInWater = %d (%sin water)\n", tr.fInWater, (tr.fInWater > 0 ? "" : "NOT "));
   UTIL_ServerConsole_printf ("   flFraction = %.2f (%d%% completed)\n", tr.flFraction, (int) tr.flFraction * 100);
   UTIL_ServerConsole_printf ("   vecEndPos = Vector (%.1f, %.1f, %.1f)\n", tr.vecEndPos.x, tr.vecEndPos.y, tr.vecEndPos.z);
   UTIL_ServerConsole_printf ("   flPlaneDist = %.2f\n", tr.flPlaneDist);
   UTIL_ServerConsole_printf ("   vecPlaneNormal = Vector (%.1f, %.1f, %.1f)\n", tr.vecPlaneNormal.x, tr.vecPlaneNormal.y, tr.vecPlaneNormal.z);
   UTIL_ServerConsole_printf ("   pHit = Entity #%d (\"%s\")\n", ENTINDEX (tr.pHit), STRING (tr.pHit->v.classname));
   UTIL_ServerConsole_printf ("   iHitgroup = %d (%s%d)\n", tr.iHitgroup, (tr.iHitgroup != 0 ? "body part #" : "nothing - "), (tr.iHitgroup != 0 ? tr.iHitgroup : 0));
   UTIL_ServerConsole_printf ("Distance traced: %.1f\n", distance * tr.flFraction);

   if (args[0][0] == 0)
   {
      UTIL_ServerConsole_printf ("Hint: you can specify ignore_monsters and ignore_glass\n");
      UTIL_ServerConsole_printf ("The syntax is: %s [distance] [ignore_monsters] [ignore_glass]\n", CMD_ARGV (0));
   }

   return;
}


void ServerCommand_TraceHull (void)
{
   float distance = 10000;
   TraceResult tr;
   char args[128][4]; // no way, I hate pointers...
   int arg_num;
   int ignore_value = 0;
   int hull_type = head_hull;

   // if no local client
   if (FNullEnt (pListenserverEntity))
   {
      UTIL_ServerConsole_printf ("%s: only works from a local client console\n", CMD_ARGV (0));
      return;
   }

   UTIL_ServerConsole_printf ("Tracing a HULL %.0f units forward from player's view angles (%.1f %.1f %.1f)\n", distance, pListenserverEntity->v.v_angle.x, pListenserverEntity->v.v_angle.y, pListenserverEntity->v.v_angle.z);

   // record arguments
   for (arg_num = 0; arg_num < 4; arg_num++)
      sprintf (args[arg_num], CMD_ARGV (arg_num + 1));

   // parse arguments
   for (arg_num = 0; arg_num < 4; arg_num++)
   {
      if ((atoi (args[arg_num]) > 0) && (atoi (args[arg_num]) <= 10000))
         distance = atoi (args[arg_num]);
      else if (strcmp ("ignore_monsters", args[arg_num]) == 0)
      {
         ignore_value |= 0x0001;
         UTIL_ServerConsole_printf ("IGNORING monsters\n");
      }
      else if (strcmp ("ignore_glass", args[arg_num]) == 0)
      {
         ignore_value |= 0x0101;
         UTIL_ServerConsole_printf ("IGNORING glass\n");
      }
      else if (strcmp ("point_hull", args[arg_num]) == 0)
      {
         hull_type = point_hull;
         UTIL_ServerConsole_printf ("Hull type: point_hull (= traceline, no size)\n");
      }
      else if (strcmp ("human_hull", args[arg_num]) == 0)
      {
         hull_type = human_hull;
         UTIL_ServerConsole_printf ("Hull type: human_hull (standing player size)\n");
      }
      else if (strcmp ("large_hull", args[arg_num]) == 0)
      {
         hull_type = large_hull;
         UTIL_ServerConsole_printf ("Hull type: large_hull (big monster size)\n");
      }
      else if (strcmp ("head_hull", args[arg_num]) == 0)
      {
         hull_type = head_hull;
         UTIL_ServerConsole_printf ("Hull type: head_hull (crouching player size)\n");
      }
   }

   MAKE_VECTORS (pListenserverEntity->v.v_angle);
   TRACE_HULL (pListenserverEntity->v.origin + pListenserverEntity->v.view_ofs,
               pListenserverEntity->v.origin + pListenserverEntity->v.view_ofs + gpGlobals->v_forward * distance,
               ignore_value, hull_type, pListenserverEntity, &tr);

   UTIL_DrawBeam (pListenserverEntity->v.origin + pListenserverEntity->v.view_ofs, pListenserverEntity->v.origin + pListenserverEntity->v.view_ofs + gpGlobals->v_forward * distance * tr.flFraction, 600, 10, 2, 250, 250, 250, 200, 10);

   UTIL_ServerConsole_printf ("Detailed trace results:\n");
   UTIL_ServerConsole_printf ("   fAllSolid = %d (plane is %svalid)\n", tr.fAllSolid, (tr.fAllSolid > 0 ? "NOT " : ""));
   UTIL_ServerConsole_printf ("   fStartSolid = %d (initial point was %sin a solid area)\n", tr.fStartSolid, (tr.fStartSolid > 0 ? "" : "NOT "));
   UTIL_ServerConsole_printf ("   fInOpen = %d (%sin open space)\n", tr.fInOpen, (tr.fInOpen > 0 ? "" : "NOT "));
   UTIL_ServerConsole_printf ("   fInWater = %d (%sin water)\n", tr.fInWater, (tr.fInWater > 0 ? "" : "NOT "));
   UTIL_ServerConsole_printf ("   flFraction = %.2f (%d%% completed)\n", tr.flFraction, (int) tr.flFraction * 100);
   UTIL_ServerConsole_printf ("   vecEndPos = Vector (%.1f, %.1f, %.1f)\n", tr.vecEndPos.x, tr.vecEndPos.y, tr.vecEndPos.z);
   UTIL_ServerConsole_printf ("   flPlaneDist = %.2f\n", tr.flPlaneDist);
   UTIL_ServerConsole_printf ("   vecPlaneNormal = Vector (%.1f, %.1f, %.1f)\n", tr.vecPlaneNormal.x, tr.vecPlaneNormal.y, tr.vecPlaneNormal.z);
   UTIL_ServerConsole_printf ("   pHit = Entity #%d (\"%s\")\n", ENTINDEX (tr.pHit), STRING (tr.pHit->v.classname));
   UTIL_ServerConsole_printf ("   iHitgroup = %d (%s%d)\n", tr.iHitgroup, (tr.iHitgroup != 0 ? "body part #" : "nothing - "), (tr.iHitgroup != 0 ? tr.iHitgroup : 0));
   UTIL_ServerConsole_printf ("Distance traced: %.1f\n", distance * tr.flFraction);

   if (args[0][0] == 0)
   {
      UTIL_ServerConsole_printf ("Hint: you can specify ignore_monsters, ignore_glass and the hull type\n");
      UTIL_ServerConsole_printf ("The syntax is: %s [distance] [ignore_monsters] [ignore_glass] [hull_type]\n", CMD_ARGV (0));
   }

   return;
}


void ServerCommand_TraceTexture (void)
{
   float distance = 10000;
   const char *texture_name;

   // if no local client
   if (FNullEnt (pListenserverEntity))
   {
      UTIL_ServerConsole_printf ("%s: only works from a local client console\n", CMD_ARGV (0));
      return;
   }

   UTIL_ServerConsole_printf ("Tracing for a TEXTURE NAME forward from player's view angles (%.1f %.1f %.1f)\n", distance, pListenserverEntity->v.v_angle.x, pListenserverEntity->v.v_angle.y, pListenserverEntity->v.v_angle.z);

   MAKE_VECTORS (pListenserverEntity->v.v_angle);
   texture_name = TRACE_TEXTURE (INDEXENT (0), pListenserverEntity->v.origin + pListenserverEntity->v.view_ofs, gpGlobals->v_forward * 9999);

   UTIL_ServerConsole_printf ("Texture name: \"%s\" (type '%c')\n", texture_name, MDLL_PM_FindTextureType ((char *) texture_name));
   return;
}


void ServerCommand_PointContents (void)
{
   int contents;

   // if no local client
   if (FNullEnt (pListenserverEntity))
   {
      UTIL_ServerConsole_printf ("%s: only works from a local client console\n", CMD_ARGV (0));
      return;
   }

   UTIL_ServerConsole_printf ("Pointing for CONTENTS at player's location (%.1f %.1f %.1f)\n", pListenserverEntity->v.origin.x, pListenserverEntity->v.origin.y, pListenserverEntity->v.origin.z);

   contents = POINT_CONTENTS (pListenserverEntity->v.origin);

   UTIL_ServerConsole_printf ("Contents: CONTENT_%s (%d)\n",
                              (contents == CONTENTS_EMPTY ? "EMPTY [air, reachable]" : 
                               (contents == CONTENTS_SOLID ? "SOLID [void / outside of map]" : 
                                (contents == CONTENTS_WATER ? "WATER" : 
                                 (contents == CONTENTS_SLIME ? "SLIME" : 
                                  (contents == CONTENTS_LAVA ? "LAVA" : 
                                   (contents == CONTENTS_SKY ? "SKY [skybox / outside of map]" : 
                                    (contents == -7 ? "ORIGIN" : 
                                     (contents == -8 ? "CLIP" : 
                                      (contents == -9 ? "CURRENT_0" : 
                                       (contents == -10 ? "CURRENT_90" : 
                                        (contents == -11 ? "CURRENT_180" : 
                                         (contents == -12 ? "CURRENT_270" : 
                                          (contents == -13 ? "CURRENT_UP" : 
                                           (contents == -14 ? "CURRENT_DOWN" : 
                                            (contents == -15 ? "TRANSLUCENT" : 
                                             (contents == CONTENTS_LADDER ? "LADDER" : 
                                              (contents == CONTENT_FLYFIELD ? "FLYFIELD" : 
                                               (contents == CONTENT_GRAVITY_FLYFIELD ? "GRAVITY_FLYFIELD" : 
                                                (contents == CONTENT_FOG ? "FOG" : 
                                                 "??? [unknown content]"))))))))))))))))))),
                              contents);
   return;
}


void ServerCommand_Locate (void)
{
   Vector v_location;

   // if no local client
   if (FNullEnt (pListenserverEntity))
   {
      UTIL_ServerConsole_printf ("%s: only works from a local client console\n", CMD_ARGV (0));
      return;
   }

   // get the desired location and display a beam towards it
   v_location.x = atof (CMD_ARGV (1));
   v_location.y = atof (CMD_ARGV (2));
   v_location.z = atof (CMD_ARGV (3));

   UTIL_ServerConsole_printf ("Locating (%.1f, %.1f, %.1f) position relatively to player for 10 seconds\n", v_location.x, v_location.y, v_location.z);
   UTIL_DrawBeam (pListenserverEntity->v.origin, v_location, 100, 10, 0, 255, 255, 255, 255, 255);
   return;
}


void ServerCommand_Teleport (void)
{
   Vector v_location;

   // if no local client
   if (FNullEnt (pListenserverEntity))
   {
      UTIL_ServerConsole_printf ("%s: only works from a local client console\n", CMD_ARGV (0));
      return;
   }

   // get the desired location and teleport player to it
   v_location.x = atof (CMD_ARGV (1));
   v_location.y = atof (CMD_ARGV (2));
   v_location.z = atof (CMD_ARGV (3));

   UTIL_ServerConsole_printf ("Teleporting player to (%.1f, %.1f, %.1f)\n", v_location.x, v_location.y, v_location.z);
   SET_ORIGIN (pListenserverEntity, v_location);
   return;
}


void ServerCommand_DumpGlobals (void)
{
   UTIL_ServerConsole_printf ("Global game variables (globalvars_t):\n");
   UTIL_ServerConsole_printf ("gpGlobals->time = %.3f\n", gpGlobals->time);
   UTIL_ServerConsole_printf ("gpGlobals->frametime = %f\n", gpGlobals->frametime);
   UTIL_ServerConsole_printf ("gpGlobals->force_retouch = %.3f\n", gpGlobals->force_retouch);
   UTIL_ServerConsole_printf ("gpGlobals->mapname = 0x%x (%s)\n", gpGlobals->mapname, STRING (gpGlobals->mapname));
   UTIL_ServerConsole_printf ("gpGlobals->startspot = 0x%x (%s)\n", gpGlobals->startspot, STRING (gpGlobals->startspot));
   UTIL_ServerConsole_printf ("gpGlobals->deathmatch = %.3f\n", gpGlobals->deathmatch);
   UTIL_ServerConsole_printf ("gpGlobals->coop = %.3f\n", gpGlobals->coop);
   UTIL_ServerConsole_printf ("gpGlobals->teamplay = %.3f\n", gpGlobals->teamplay);
   UTIL_ServerConsole_printf ("gpGlobals->serverflags = %f\n", gpGlobals->serverflags);
   UTIL_ServerConsole_printf ("gpGlobals->found_secrets = %.3f\n", gpGlobals->found_secrets);
   UTIL_ServerConsole_printf ("gpGlobals->v_forward = Vector (%.1f, %.1f, %.1f)\n", gpGlobals->v_forward.x, gpGlobals->v_forward.y, gpGlobals->v_forward.z);
   UTIL_ServerConsole_printf ("gpGlobals->v_up = Vector (%.1f, %.1f, %.1f)\n", gpGlobals->v_up.x, gpGlobals->v_up.y, gpGlobals->v_up.z);
   UTIL_ServerConsole_printf ("gpGlobals->v_right = Vector (%.1f, %.1f, %.1f)\n", gpGlobals->v_right.x, gpGlobals->v_right.y, gpGlobals->v_right.z);
   UTIL_ServerConsole_printf ("gpGlobals->trace_allsolid = %.3f\n", gpGlobals->trace_allsolid);
   UTIL_ServerConsole_printf ("gpGlobals->trace_startsolid = %.3f\n", gpGlobals->trace_startsolid);
   UTIL_ServerConsole_printf ("gpGlobals->trace_fraction = %.3f\n", gpGlobals->trace_fraction);
   UTIL_ServerConsole_printf ("gpGlobals->trace_endpos = Vector (%.1f, %.1f, %.1f)\n", gpGlobals->trace_endpos.x, gpGlobals->trace_endpos.y, gpGlobals->trace_endpos.z);
   UTIL_ServerConsole_printf ("gpGlobals->trace_plane_normal = Vector (%.1f, %.1f, %.1f)\n", gpGlobals->trace_plane_normal.x, gpGlobals->trace_plane_normal.y, gpGlobals->trace_plane_normal.z);
   UTIL_ServerConsole_printf ("gpGlobals->trace_plane_dist = %.3f\n", gpGlobals->trace_plane_dist);
   if (!FNullEnt (gpGlobals->trace_ent))
      UTIL_ServerConsole_printf ("gpGlobals->trace_ent = Entity #%d (\"%s\")\n", ENTINDEX (gpGlobals->trace_ent), STRING (gpGlobals->trace_ent->v.classname));
   else
      UTIL_ServerConsole_printf ("gpGlobals->trace_ent = 0x%x (NULLENT)\n", gpGlobals->trace_ent); 
   UTIL_ServerConsole_printf ("gpGlobals->trace_inopen = %.3f\n", gpGlobals->trace_inopen);
   UTIL_ServerConsole_printf ("gpGlobals->trace_inwater = %.3f\n", gpGlobals->trace_inwater);
   UTIL_ServerConsole_printf ("gpGlobals->trace_hitgroup = %d (%s%d)\n", gpGlobals->trace_hitgroup, (gpGlobals->trace_hitgroup != 0 ? "body part #" : "nothing - "), (gpGlobals->trace_hitgroup != 0 ? gpGlobals->trace_hitgroup : 0));
   UTIL_ServerConsole_printf ("gpGlobals->trace_flags = %d (%s)\n", gpGlobals->trace_flags, (gpGlobals->trace_flags & FTRACE_SIMPLEBOX ? "FTRACE_SIMPLEBOX" : "none"));
   UTIL_ServerConsole_printf ("gpGlobals->msg_entity = %d (destination of message is entity #%d - \"%s\")\n", gpGlobals->msg_entity, gpGlobals->msg_entity, (INDEXENT (gpGlobals->msg_entity)->v.flags & FL_CLIENT ? STRING (INDEXENT (gpGlobals->msg_entity)->v.netname) : STRING (INDEXENT (gpGlobals->msg_entity)->v.classname)));
   UTIL_ServerConsole_printf ("gpGlobals->cdAudioTrack = %d\n", gpGlobals->cdAudioTrack);
   UTIL_ServerConsole_printf ("gpGlobals->maxClients = %d\n", gpGlobals->maxClients);
   UTIL_ServerConsole_printf ("gpGlobals->maxEntities = %d\n", gpGlobals->maxEntities);
   UTIL_ServerConsole_printf ("gpGlobals->pStringBase = %d (0x%x)\n", gpGlobals->pStringBase, gpGlobals->pStringBase);
   UTIL_ServerConsole_printf ("gpGlobals->pSaveData = %d (0x%x)\n", gpGlobals->pSaveData, gpGlobals->pSaveData);
   UTIL_ServerConsole_printf ("gpGlobals->vecLandmarkOffset = Vector (%.1f, %.1f, %.1f)\n", gpGlobals->vecLandmarkOffset.x, gpGlobals->vecLandmarkOffset.y, gpGlobals->vecLandmarkOffset.z);

   return;
}


void ServerCommand_Extract (void)
{
   // extracts a file from the PAK cache (also works to extract from the STEAM cache)

   char *filebuf = NULL;
   FILE *fp;
   char filename[128];
   char new_filename[128];
   int size;
   
   // no file specified ?
   if ((CMD_ARGV (1) == NULL) || (*CMD_ARGV (1) == 0))
   {
      UTIL_ServerConsole_printf ("%s: must specify file to extract.\n", CMD_ARGV (0));
      return;
   }

   // build input and output file names
   sprintf (filename, "%s", CMD_ARGV (1));
   if ((CMD_ARGV (2) == NULL) || (*CMD_ARGV (2) == 0))
      sprintf (new_filename, "%s/%s", GameDir (), CMD_ARGV (1)); // no output file name specified
   else
      sprintf (new_filename, "%s/%s", GameDir (), CMD_ARGV (2)); // output file name specified

   // attempt to open the file
   filebuf = (char *) LOAD_FILE_FOR_ME (filename, &size);

   // file not found ?
   if (filebuf == NULL)
   {
      UTIL_ServerConsole_printf ("%s: file not found in the running PAK/Steam cache.\n", filename);
      return;
   }

   // test to see if the file already exists outside the Steam cache
   fp = fopen (new_filename, "rb");
   if (fp != NULL)
   {
      fclose (fp);
      FREE_FILE (filebuf);
      UTIL_ServerConsole_printf ("%s: file already exists outside the PAK/Steam cache.\n", new_filename);
      UTIL_ServerConsole_printf ("Please specify another filename as 2nd argument.\n");
      return;
   }

   // file does not exist, create it
   fp = fopen (new_filename, "wb");
   if (fp == NULL)
   {
      FREE_FILE (filebuf);
      UTIL_ServerConsole_printf ("%s: failure creating the new file '%s' on disk.\n", CMD_ARGV (0), new_filename);
      UTIL_ServerConsole_printf ("Ensure the target directory already exists.\n");
      return;
   }

   // dump the file and close it
   fwrite (filebuf, size, 1, fp);
   fclose (fp);
   FREE_FILE (filebuf);

   UTIL_ServerConsole_printf ("File '%s' dumped from the PAK/Steam cache to disk.\n", filename);
   return;
}


void UTIL_DrawBeam (Vector start, Vector end, int life, int width, int noise, int red, int green, int blue, int brightness, int speed)
{
   if (FNullEnt (pListenserverEntity))
      return; // reliability check

   MESSAGE_BEGIN (MSG_ONE_UNRELIABLE, SVC_TEMPENTITY, NULL, pListenserverEntity);
   WRITE_BYTE (TE_BEAMPOINTS);
   WRITE_COORD (start.x);
   WRITE_COORD (start.y);
   WRITE_COORD (start.z);
   WRITE_COORD (end.x);
   WRITE_COORD (end.y);
   WRITE_COORD (end.z);
   WRITE_SHORT (beam_texture);
   WRITE_BYTE (1); // framestart
   WRITE_BYTE (10); // framerate
   WRITE_BYTE (life); // life in 0.1's
   WRITE_BYTE (width); // width
   WRITE_BYTE (noise); // noise
   WRITE_BYTE (red); // r, g, b
   WRITE_BYTE (green); // r, g, b
   WRITE_BYTE (blue); // r, g, b
   WRITE_BYTE (brightness); // brightness
   WRITE_BYTE (speed); // speed
   MESSAGE_END ();
}


void UTIL_printf (char *fmt, ...)
{
   FILE *fp;

   va_list argptr;
   static char string[1024];

   va_start (argptr, fmt);
   vsprintf (string, fmt, argptr);
   va_end (argptr);

   // if a listen server is watching, print on his HUD
   if (!FNullEnt (pListenserverEntity))
   {
      MESSAGE_BEGIN (MSG_ONE_UNRELIABLE, GET_USER_MSG_ID (PLID, "SayText", NULL), NULL, pListenserverEntity);
      WRITE_BYTE (ENTINDEX (pListenserverEntity));
      WRITE_STRING (string); // print to HUD
      MESSAGE_END ();
   }
   else
      SERVER_PRINT (string); // else print to console

   // has a log file been specified ?
   if (strcmp (CVAR_GET_STRING ("logfile"), "") != 0)
   {
      fp = fopen (CVAR_GET_STRING ("logfile"), "a"); // if so, open the file in ASCII append mode
      fprintf (fp, "PMTOOLS (HUD): %s", string); // append the string to the file
      fclose (fp); // close the file
   }

   return;
}


void UTIL_ServerConsole_printf (char *fmt, ...)
{
   FILE *fp;

   va_list argptr;
   static char string[1024];

   va_start (argptr, fmt);
   vsprintf (string, fmt, argptr);
   va_end (argptr);

   SERVER_PRINT (string); // print to console

   // has a log file been specified ?
   if (strcmp (CVAR_GET_STRING ("logfile"), "") != 0)
   {
      fp = fopen (CVAR_GET_STRING ("logfile"), "a"); // if so, open the file in ASCII append mode
      fprintf (fp, "PMTOOLS (console): %s", string); // append the string to the file
      fclose (fp); // close the file
   }
}


bool FInViewCone (Vector pOrigin, edict_t *pEdict)
{
   Vector2D vec2LOS;
   float flDot;

   if (FNullEnt (pEdict))
      return FALSE; // reliability check

   MAKE_VECTORS (pEdict->v.angles);

   vec2LOS = (pOrigin - pEdict->v.origin).Make2D ();
   vec2LOS = vec2LOS.Normalize ();

   flDot = DotProduct (vec2LOS, gpGlobals->v_forward.Make2D ());

   // 60 degree field of view 
   if (flDot > 0.50)
      return TRUE;
   else
      return FALSE;
}


bool CanSee (edict_t *pSeeingEntity, edict_t *pEntityToSee)
{
   TraceResult tr;

   if (FNullEnt (pSeeingEntity) || FNullEnt (pEntityToSee))
      return FALSE; // reliability check

   if (pEntityToSee->v.flags & EF_NODRAW)
      return FALSE; // if not drawn, then it is invisible

   // fire a trace line from source to destination
   TRACE_LINE (pSeeingEntity->v.absmin + (pSeeingEntity->v.size * 0.5) + pSeeingEntity->v.view_ofs,
               pEntityToSee->v.absmin + (pEntityToSee->v.size * 0.5),
               TRUE, pSeeingEntity, &tr);
   if ((tr.flFraction == 1.0) || (ENTINDEX (tr.pHit) == ENTINDEX (pEntityToSee)))
      return TRUE; // line of sight is valid.

   return FALSE; // line of sight is not established
}


char *GameDir (void)
{
   // This function fixes the erratic behaviour caused by the use of the GET_GAME_DIR macro.

   static char szGetGameDir[128];
   unsigned char length, fieldstart, fieldstop;

   memset (szGetGameDir, 0, sizeof (szGetGameDir));
   GET_GAME_DIR (szGetGameDir); // call the engine macro and let it mallocate for the char pointer

   length = strlen (szGetGameDir); // get the length of the returned string
   length--; // ignore the trailing string terminator

   // format the returned string to get the last directory name
   fieldstop = length;
   while (((szGetGameDir[fieldstop] == '\\') || (szGetGameDir[fieldstop] == '/')) && (fieldstop > 0))
      fieldstop--; // shift back any trailing separator

   fieldstart = fieldstop;
   while ((szGetGameDir[fieldstart] != '\\') && (szGetGameDir[fieldstart] != '/') && (fieldstart > 0))
      fieldstart--; // shift back to the start of the last subdirectory name

   if ((szGetGameDir[fieldstart] == '\\') || (szGetGameDir[fieldstart] == '/'))
      fieldstart++; // if we reached a separator, step over it

   // now copy the formatted string back onto itself character per character
   for (length = fieldstart; length <= fieldstop; length++)
      szGetGameDir[length - fieldstart] = szGetGameDir[length];
   szGetGameDir[length - fieldstart] = 0; // terminate the string

   return (&szGetGameDir[0]); // and return a pointer to it
}
